import logging
import os
import pathlib
from abc import abstractmethod
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

import dask.dataframe as dd
import pandas as pd
import pystow
from pystow.utils import read_zipfile_csv
from slugify import slugify

from .dask import read_dask_df_archive_csv
from .typing import BACKEND_LITERAL, COLUMNS, EA_SIDES
from .utils import fix_dataclass_init_docs

BASE_DATASET_KEY = "sylloge"

BASE_DATASET_MODULE = pystow.module(BASE_DATASET_KEY)

DataFrameType = TypeVar("DataFrameType", pd.DataFrame, dd.DataFrame)


if TYPE_CHECKING:
    import dask.dataframe as dd

logger = logging.getLogger(__name__)


@fix_dataclass_init_docs
@dataclass
class TrainTestValSplit(Generic[DataFrameType]):
    """Dataclass holding split of gold standard entity links."""

    #: entity links for training
    train: DataFrameType
    #: entity links for testing
    test: DataFrameType
    #: entity links for validation
    val: DataFrameType


@fix_dataclass_init_docs
class EADataset(Generic[DataFrameType]):
    """Dataset class holding information of the alignment class."""

    rel_triples_left: DataFrameType
    rel_triples_right: DataFrameType
    attr_triples_left: DataFrameType
    attr_triples_right: DataFrameType
    ent_links: DataFrameType
    dataset_names: Tuple[str, str]
    folds: Optional[Sequence[TrainTestValSplit[DataFrameType]]] = None

    _REL_TRIPLES_LEFT_PATH: str = "rel_triples_left_parquet"
    _REL_TRIPLES_RIGHT_PATH: str = "rel_triples_right_parquet"
    _ATTR_TRIPLES_LEFT_PATH: str = "attr_triples_left_parquet"
    _ATTR_TRIPLES_RIGHT_PATH: str = "attr_triples_right_parquet"
    _ENT_LINKS_PATH: str = "ent_links_parquet"
    _FOLD_DIR: str = "folds"
    _TRAIN_LINKS_PATH: str = "train_parquet"
    _TEST_LINKS_PATH: str = "test_parquet"
    _VAL_LINKS_PATH: str = "val_parquet"
    _DATASET_NAMES_PATH: str = "dataset_names.txt"

    def __init__(
        self,
        *,
        rel_triples_left: DataFrameType,
        rel_triples_right: DataFrameType,
        attr_triples_left: DataFrameType,
        attr_triples_right: DataFrameType,
        ent_links: DataFrameType,
        dataset_names: Tuple[str, str],
        folds: Optional[Sequence[TrainTestValSplit[DataFrameType]]] = None,
        backend: BACKEND_LITERAL = "pandas",
        npartitions: int = 1,
    ) -> None:
        """Create an entity aligment dataclass.

        :param rel_triples_left: relation triples of left knowledge graph
        :param rel_triples_right: relation triples of right knowledge graph
        :param attr_triples_left: attribute triples of left knowledge graph
        :param attr_triples_right: attribute triples of right knowledge graph
        :param dataset_names: tuple of dataset names
        :param ent_links: gold standard entity links of alignment
        :param folds: optional pre-split folds of the gold standard
        :param backend: which backend is used of either 'pandas' or 'dask'
        :param npartitions: how many partitions to use for each frame, when using dask
        """
        self.rel_triples_left = rel_triples_left
        self.rel_triples_right = rel_triples_right
        self.attr_triples_left = attr_triples_left
        self.attr_triples_right = attr_triples_right
        self.ent_links = ent_links
        self.dataset_names = dataset_names
        self.folds = folds
        self.npartitions: int = npartitions
        self._backend: BACKEND_LITERAL = backend
        # trigger possible transformation
        self.backend = backend

    @property
    def _canonical_name(self) -> str:
        raise NotImplementedError

    @property
    def canonical_name(self) -> str:
        """A canonical name for this dataset instance.

        This includes all the necessary information
        to distinguish this specific dataset as string.
        This can be used e.g. to create folders with this
        dataset name to store results.

        :return: concise string representation for this dataset instance
        """
        name = self._canonical_name
        assert isinstance(name, str)  # for mypy
        return slugify(name, separator="_")

    @property
    def _param_repr(self) -> str:
        raise NotImplementedError

    @property
    def _statistics(self) -> str:
        if isinstance(self.rel_triples_left, pd.DataFrame):
            return f"rel_triples_left={len(self.rel_triples_left)}, rel_triples_right={len(self.rel_triples_right)}, attr_triples_left={len(self.attr_triples_left)}, attr_triples_right={len(self.attr_triples_right)}, ent_links={len(self.ent_links)}, folds={len(self.folds) if self.folds else None}"  # type: ignore
        else:
            unknown = "unknown_len"
            return f"rel_triples_left={unknown}, rel_triples_right={unknown}, attr_triples_left={unknown}, attr_triples_right={unknown}, ent_links={unknown}, folds={unknown if self.folds else None}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(backend={self.backend}, {self._param_repr}{self._statistics})"

    def _additional_backend_handling(self, backend: BACKEND_LITERAL):
        pass

    @property
    def backend(self) -> BACKEND_LITERAL:
        return self._backend

    @backend.setter
    def backend(self, backend: BACKEND_LITERAL):
        """Set backend and transform data if needed"""
        if backend == "pandas":
            self._backend = "pandas"
            if isinstance(self.rel_triples_left, pd.DataFrame):
                return
            else:
                self.rel_triples_left = self.rel_triples_left.compute()
                self.rel_triples_right = self.rel_triples_right.compute()
                self.attr_triples_left = self.attr_triples_left.compute()
                self.attr_triples_right = self.attr_triples_right.compute()
                self.ent_links = self.ent_links.compute()
                if self.folds:
                    for fold in self.folds:
                        fold.train = fold.train.compute()
                        fold.test = fold.test.compute()
                        fold.val = fold.val.compute()

        elif backend == "dask":
            self._backend = "dask"
            if isinstance(self.rel_triples_left, dd.DataFrame):
                if self.rel_triples_left.npartitions != self.npartitions:
                    self.rel_triples_left = self.rel_triples_left.repartition(
                        npartitions=self.npartitions
                    )
                    self.rel_triples_right = self.rel_triples_right.repartition(
                        npartitions=self.npartitions
                    )
                    self.attr_triples_left = self.attr_triples_left.repartition(
                        npartitions=self.npartitions
                    )
                    self.attr_triples_right = self.attr_triples_right.repartition(
                        npartitions=self.npartitions
                    )
                    self.ent_links = self.ent_links.repartition(
                        npartitions=self.npartitions
                    )
                    if self.folds:
                        for fold in self.folds:
                            fold.train = fold.train.repartition(
                                npartitions=self.npartitions
                            )
                            fold.test = fold.test.repartition(
                                npartitions=self.npartitions
                            )
                            fold.val = fold.val.repartition(
                                npartitions=self.npartitions
                            )
                else:
                    return

            else:
                self.rel_triples_left = dd.from_pandas(
                    self.rel_triples_left, npartitions=self.npartitions
                )
                self.rel_triples_right = dd.from_pandas(
                    self.rel_triples_right, npartitions=self.npartitions
                )
                self.attr_triples_left = dd.from_pandas(
                    self.attr_triples_left, npartitions=self.npartitions
                )
                self.attr_triples_right = dd.from_pandas(
                    self.attr_triples_right, npartitions=self.npartitions
                )
                self.ent_links = dd.from_pandas(
                    self.ent_links, npartitions=self.npartitions
                )
                if self.folds:
                    for fold in self.folds:
                        fold.train = dd.from_pandas(
                            fold.train, npartitions=self.npartitions
                        )
                        fold.test = dd.from_pandas(
                            fold.test, npartitions=self.npartitions
                        )
                        fold.val = dd.from_pandas(
                            fold.val, npartitions=self.npartitions
                        )
        else:
            raise ValueError(f"Unknown backend {backend}")
        self._additional_backend_handling(backend)

    def to_parquet(self, path: Union[str, pathlib.Path], **kwargs):
        """Write dataset to path as several parquet files.

        :param path: directory where dataset will be stored. Will be created if necessary.
        :param kwargs: will be handed through to `to_parquet` functions

        .. seealso:: :func:`read_parquet`
        """
        if not os.path.exists(path):
            os.makedirs(path)
        if not isinstance(path, pathlib.Path):
            path = pathlib.Path(path)

        # write dataset names
        with open(path.joinpath(self.__class__._DATASET_NAMES_PATH), "w") as fh:
            for side, name in zip(EA_SIDES, self.dataset_names):
                fh.write(f"{side}:{name}\n")

        # write tables
        for table, table_path in zip(
            [
                self.rel_triples_left,
                self.rel_triples_right,
                self.attr_triples_left,
                self.attr_triples_right,
                self.ent_links,
            ],
            [
                self.__class__._REL_TRIPLES_LEFT_PATH,
                self.__class__._REL_TRIPLES_RIGHT_PATH,
                self.__class__._ATTR_TRIPLES_LEFT_PATH,
                self.__class__._ATTR_TRIPLES_RIGHT_PATH,
                self.__class__._ENT_LINKS_PATH,
            ],
        ):
            table.to_parquet(path.joinpath(table_path), **kwargs)

        # write folds
        if self.folds:
            fold_path = path.joinpath(self.__class__._FOLD_DIR)
            for fold_number, fold in enumerate(self.folds, start=1):
                fold_dir = fold_path.joinpath(str(fold_number))
                os.makedirs(fold_dir)
                for links, link_path in zip(
                    [fold.train, fold.test, fold.val],
                    [
                        self.__class__._TRAIN_LINKS_PATH,
                        self.__class__._TEST_LINKS_PATH,
                        self.__class__._VAL_LINKS_PATH,
                    ],
                ):
                    table.to_parquet(fold_dir.joinpath(link_path), **kwargs)

    @staticmethod
    def _read_parquet_values(
        path: Union[str, pathlib.Path], backend: BACKEND_LITERAL = "pandas", **kwargs
    ) -> Dict[str, Any]:
        if not isinstance(path, pathlib.Path):
            path = pathlib.Path(path)

        read_parquet_fn = pd.read_parquet if backend == "pandas" else dd.read_parquet

        # read dataset names
        with open(path.joinpath(EADataset._DATASET_NAMES_PATH), "r") as fh:
            dataset_names = tuple(line.strip().split(":")[1] for line in fh)
            # for mypy
            dataset_names = cast(Tuple[str, str], dataset_names)

        tables = {}
        # read tables
        for table, table_path in zip(
            [
                "rel_triples_left",
                "rel_triples_right",
                "attr_triples_left",
                "attr_triples_right",
                "ent_links",
            ],
            [
                EADataset._REL_TRIPLES_LEFT_PATH,
                EADataset._REL_TRIPLES_RIGHT_PATH,
                EADataset._ATTR_TRIPLES_LEFT_PATH,
                EADataset._ATTR_TRIPLES_RIGHT_PATH,
                EADataset._ENT_LINKS_PATH,
            ],
        ):
            tables[table] = read_parquet_fn(path.joinpath(table_path), **kwargs)

        # read folds
        fold_path = path.joinpath(EADataset._FOLD_DIR)
        folds = None
        if os.path.exists(fold_path):
            folds = []
            for tmp_fold_dir in sorted(sub_dir for sub_dir in os.listdir(fold_path)):
                fold_dir = fold_path.joinpath(tmp_fold_dir)
                train_test_val = {}
                for links, link_path in zip(
                    ["train", "test", "val"],
                    [
                        EADataset._TRAIN_LINKS_PATH,
                        EADataset._TEST_LINKS_PATH,
                        EADataset._VAL_LINKS_PATH,
                    ],
                ):
                    train_test_val[links] = read_parquet_fn(
                        fold_dir.joinpath(link_path), **kwargs
                    )
                folds.append(TrainTestValSplit(**train_test_val))
        npartitions = 1
        if backend == "dask":
            npartitions = tables["rel_triples_left"].npartitions
        return dict(
            dataset_names=dataset_names,
            folds=folds,
            backend=backend,
            npartitions=npartitions,
            **tables,
        )

    @classmethod
    def read_parquet(
        cls,
        path: Union[str, pathlib.Path],
        backend: BACKEND_LITERAL = "pandas",
        **kwargs,
    ) -> "EADataset":
        """Read dataset from parquet files in given `path`.

        This function expects the left/right attribute/relation triples and entity links as well as a `dataset_names.txt`

        Optionally folds are read from a `folds` directory, with numbered fold subdirectories containing train/test/val links.

        :param path: Directory with files
        :param backend: Whether to use pandas or dask for reading
        :param kwargs: passed on to the respective read function
        :return: EADataset read from parquet

        .. seealso:: :func:`to_parquet`
        """
        return cls(
            **EADataset._read_parquet_values(path=path, backend=backend, **kwargs)
        )


class CacheableEADataset(EADataset):
    def __init__(
        self,
        *,
        cache_path: pathlib.Path,
        use_cache: bool = True,
        parquet_load_options: Optional[Mapping] = None,
        parquet_store_options: Optional[Mapping] = None,
        **init_kwargs,
    ):
        """EADataset that uses caching after initial read.

        :param cache_path: Path where cache will be stored/loaded
        :param use_cache: whether to use cache
        :param parquet_load_options: handed through to parquet loading function
        :param parquet_store_options: handed through to parquet writing function
        :param init_kwargs: other arguments for creating the EADataset instance
        """
        self.cache_path = cache_path
        self.parquet_load_options = parquet_load_options or {}
        self.parquet_store_options = parquet_store_options or {}
        backend = init_kwargs["backend"]
        specific_npartitions = init_kwargs["npartitions"]
        update_cache = False
        if use_cache:
            if self.cache_path.exists():
                logger.info(f"Loading from cache at {self.cache_path}")
                init_kwargs.update(self.load_from_cache(backend=backend))
            else:
                init_kwargs.update(self.initial_read(backend=backend))
                update_cache = True
        else:
            init_kwargs.update(self.initial_read(backend=backend))
        if specific_npartitions != 1:
            init_kwargs["npartitions"] = specific_npartitions
        super().__init__(**init_kwargs)
        if update_cache:
            logger.info(f"Caching dataset at {self.cache_path}")
            self.store_cache()

    def create_cache_path(
        self,
        pystow_module: pystow.Module,
        inner_cache_path: str,
        cache_path: Optional[pathlib.Path] = None,
    ) -> pathlib.Path:
        """Uses either pystow module or cache_path to create cache path.

        :param pystow_module: module where data is stored
        :param inner_cache_path: path relative to pystow/cache path
        :param cache_path: alternative to pystow module
        :return: cache path as `pathlib.Path`
        """
        if cache_path is None:
            return pystow_module.join("cached", inner_cache_path, ensure_exists=False)
        else:
            return cache_path.joinpath(inner_cache_path)

    def load_from_cache(self, backend: BACKEND_LITERAL = "pandas") -> Dict[str, Any]:
        return EADataset._read_parquet_values(path=self.cache_path, backend=backend)

    @abstractmethod
    def initial_read(self, backend: BACKEND_LITERAL) -> Dict[str, Any]:
        """Read data for initialising EADataset."""

    def store_cache(self):
        self.to_parquet(self.cache_path, **self.parquet_store_options)


class ZipEADataset(CacheableEADataset):
    """Dataset created from zip file which is downloaded."""

    def __init__(
        self,
        *,
        cache_path: pathlib.Path,
        zip_path: str,
        inner_path: pathlib.PurePosixPath,
        dataset_names: Tuple[str, str],
        file_name_rel_triples_left: str = "rel_triples_1",
        file_name_rel_triples_right: str = "rel_triples_2",
        file_name_attr_triples_left: str = "attr_triples_1",
        file_name_attr_triples_right: str = "attr_triples_2",
        file_name_ent_links: str = "ent_links",
        backend: BACKEND_LITERAL = "pandas",
        npartitions: int = 1,
        use_cache: bool = True,
    ):
        """Initialize ZipEADataset.

        :param cache_path: Path where cache will be stored/loaded
        :param zip_path: path to zip archive containing data
        :param inner_path: base path inside zip archive
        :param dataset_names: tuple of dataset names
        :param file_name_rel_triples_left: file name of left relation triples
        :param file_name_rel_triples_right: file name of right relation triples
        :param file_name_attr_triples_left: file name of left attribute triples
        :param file_name_attr_triples_right: file name of right attribute triples
        :param file_name_ent_links: file name gold standard containing all entity links
        :param backend: Whether to use "pandas" or "dask"
        :param npartitions: how many partitions to use for each frame, when using dask
        :param use_cache: whether to use cache or not
        """
        self.zip_path = zip_path
        self.inner_path = inner_path
        self.file_name_rel_triples_left = file_name_rel_triples_left
        self.file_name_rel_triples_right = file_name_rel_triples_right
        self.file_name_ent_links = file_name_ent_links
        self.file_name_attr_triples_left = file_name_attr_triples_left
        self.file_name_attr_triples_right = file_name_attr_triples_right

        super().__init__(
            dataset_names=dataset_names,
            cache_path=cache_path,
            backend=backend,
            npartitions=npartitions,
            use_cache=use_cache,
        )

    def initial_read(self, backend: BACKEND_LITERAL) -> Dict[str, Any]:
        return dict(
            rel_triples_left=self._read_triples(
                file_name=self.file_name_rel_triples_left, backend=backend
            ),
            rel_triples_right=self._read_triples(
                file_name=self.file_name_rel_triples_right, backend=backend
            ),
            attr_triples_left=self._read_triples(
                file_name=self.file_name_attr_triples_left, backend=backend
            ),
            attr_triples_right=self._read_triples(
                file_name=self.file_name_attr_triples_right, backend=backend
            ),
            ent_links=self._read_triples(
                file_name=self.file_name_ent_links, is_links=True, backend=backend
            ),
        )

    @overload
    def _read_triples(
        self,
        file_name: Union[str, pathlib.Path],
        backend: Literal["pandas"],
        is_links: bool = False,
    ) -> pd.DataFrame:
        ...

    @overload
    def _read_triples(
        self,
        file_name: Union[str, pathlib.Path],
        backend: Literal["dask"],
        is_links: bool = False,
    ) -> "dd.DataFrame":
        ...

    def _read_triples(
        self,
        file_name: Union[str, pathlib.Path],
        backend: BACKEND_LITERAL,
        is_links: bool = False,
    ) -> Union[pd.DataFrame, "dd.DataFrame"]:
        columns = list(EA_SIDES) if is_links else COLUMNS
        read_csv_kwargs = dict(
            header=None,
            names=columns,
            sep="\t",
            encoding="utf8",
            dtype=str,
        )
        if backend == "pandas":
            return read_zipfile_csv(
                path=self.zip_path,
                inner_path=str(self.inner_path.joinpath(file_name)),
                **read_csv_kwargs,
            )
        else:
            return read_dask_df_archive_csv(
                path=self.zip_path,
                inner_path=str(self.inner_path.joinpath(file_name)),
                protocol="zip",
                **read_csv_kwargs,
            )


class ZipEADatasetWithPreSplitFolds(ZipEADataset):
    """Dataset with pre-split folds created from zip file which is downloaded."""

    def __init__(
        self,
        *,
        cache_path: pathlib.Path,
        zip_path: str,
        inner_path: pathlib.PurePosixPath,
        dataset_names: Tuple[str, str],
        file_name_rel_triples_left: str = "rel_triples_1",
        file_name_rel_triples_right: str = "rel_triples_2",
        file_name_ent_links: str = "ent_links",
        file_name_attr_triples_left: str = "attr_triples_1",
        file_name_attr_triples_right: str = "attr_triples_2",
        backend: BACKEND_LITERAL = "pandas",
        npartitions: int = 1,
        directory_name_folds: str = "721_5fold",
        directory_names_individual_folds: Sequence[str] = ("1", "2", "3", "4", "5"),
        file_name_test_links: str = "test_links",
        file_name_train_links: str = "train_links",
        file_name_valid_links: str = "valid_links",
        use_cache: bool = True,
    ):
        """Initialize ZipEADatasetWithPreSplitFolds.

        :param cache_path: Path where cache will be stored/loaded
        :param zip_path: path to zip archive containing data
        :param inner_path: base path inside zip archive
        :param dataset_names: tuple of dataset names
        :param file_name_rel_triples_left: file name of left relation triples
        :param file_name_rel_triples_right: file name of right relation triples
        :param file_name_attr_triples_left: file name of left attribute triples
        :param file_name_attr_triples_right: file name of right attribute triples
        :param file_name_ent_links: file name gold standard containing all entity links
        :param backend: Whether to use "pandas" or "dask"
        :param npartitions: how many partitions to use for each frame, when using dask
        :param directory_name_folds: name of the folds directory
        :param directory_names_individual_folds: name of individual folds
        :param file_name_test_links: name of test link file
        :param file_name_train_links: name of train link file
        :param file_name_valid_links: name of valid link file
        :param use_cache: whether to use cache or not
        """
        self.zip_path = zip_path
        self.inner_path = inner_path
        self.directory_names_individual_folds = directory_names_individual_folds
        self.directory_name_folds = directory_name_folds
        self.file_name_train_links = file_name_train_links
        self.file_name_test_links = file_name_test_links
        self.file_name_valid_links = file_name_valid_links

        super().__init__(
            dataset_names=dataset_names,
            zip_path=zip_path,
            inner_path=inner_path,
            cache_path=cache_path,
            backend=backend,
            npartitions=npartitions,
            use_cache=use_cache,
            file_name_rel_triples_left=file_name_rel_triples_left,
            file_name_rel_triples_right=file_name_rel_triples_right,
            file_name_ent_links=file_name_ent_links,
            file_name_attr_triples_left=file_name_attr_triples_left,
            file_name_attr_triples_right=file_name_attr_triples_right,
        )

    def initial_read(self, backend: BACKEND_LITERAL):
        folds = []
        for fold in self.directory_names_individual_folds:
            fold_folder = pathlib.Path(self.directory_name_folds).joinpath(fold)
            train = self._read_triples(
                fold_folder.joinpath(self.file_name_train_links),
                is_links=True,
                backend=backend,
            )
            test = self._read_triples(
                fold_folder.joinpath(self.file_name_test_links),
                is_links=True,
                backend=backend,
            )
            val = self._read_triples(
                fold_folder.joinpath(self.file_name_valid_links),
                is_links=True,
                backend=backend,
            )
            folds.append(TrainTestValSplit(train=train, test=test, val=val))
        return {**super().initial_read(backend=backend), **dict(folds=folds)}



"""OAEI knowledge graph track dataset class."""
import logging
import pathlib
import typing
from collections import namedtuple
from typing import Dict, Literal, Optional, Tuple

import dask.dataframe as dd
import pandas as pd

from .base import BASE_DATASET_MODULE, CacheableEADataset
from .dask import read_dask_bag_from_archive_text
from .typing import (
    BACKEND_LITERAL,
    COLUMNS,
    EA_SIDE_LEFT,
    EA_SIDE_RIGHT,
    LABEL_HEAD,
    LABEL_RELATION,
    LABEL_TAIL,
)

OAEI_MODULE = BASE_DATASET_MODULE.module("oaei")

OAEI_TASK_NAME = Literal[
    "starwars-swg",
    "starwars-swtor",
    "marvelcinematicuniverse-marvel",
    "memoryalpha-memorybeta",
    "memoryalpha-stexpanded",
]

logger = logging.getLogger(__name__)

REFLINE_TYPE_LITERAL = Literal["dismiss", "entity", "property", "class", "unknown"]
REL_ATTR_LITERAL = Literal["rel", "attr"]

reflinetype = pd.CategoricalDtype(categories=typing.get_args(REFLINE_TYPE_LITERAL))
relattrlinetype = pd.CategoricalDtype(categories=typing.get_args(REL_ATTR_LITERAL))

URL_SHA512_HASH = namedtuple("URL_SHA512_HASH", ["url", "hash"])


def parse_ref_line(line: str) -> Tuple[str, str, str, REFLINE_TYPE_LITERAL]:
    subj, pred, obj = line.split(" ", maxsplit=2)
    pred = pred[1:-1]
    obj = obj[1:-4]
    disallowed_pred = [
        "http://knowledgeweb.semanticweb.org/heterogeneity/alignmentmap",
        "http://knowledgeweb.semanticweb.org/heterogeneity/alignmentrelation",
        "http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
        "http://knowledgeweb.semanticweb.org/heterogeneity/alignmentmeasure",
        "http://knowledgeweb.semanticweb.org/heterogeneity/alignmenttype",
        "http://knowledgeweb.semanticweb.org/heterogeneity/alignmentlevel",
        "http://knowledgeweb.semanticweb.org/heterogeneity/alignmentxml",
    ]
    line_type: REFLINE_TYPE_LITERAL = "unknown"
    if pred in disallowed_pred:
        line_type = "dismiss"
    if "/resource/" in obj:
        line_type = "entity"
    elif "/property/" in obj:
        line_type = "property"
    elif "/class/" in obj:
        line_type = "class"
    return subj, pred, obj, line_type


def clean_and_categorize_part(
    part: str, remove_dt: bool = True
) -> Tuple[str, REL_ATTR_LITERAL]:
    if part.endswith(" .\n"):
        part = part[:-3]
    if part.startswith("<") and part.endswith(">"):
        return part[1:-1], "rel"
    if remove_dt:
        part = part.split("^^")[0]
        if part.startswith('"') and part.endswith('"'):
            return part[1:-1], "attr"
    return part, "attr"


def fault_tolerant_parse_nt(
    line: str, remove_dt: bool = True
) -> Tuple[str, str, str, REL_ATTR_LITERAL]:
    subj, pred, obj = line.split(" ", maxsplit=2)
    subj, _ = clean_and_categorize_part(subj, remove_dt)
    pred, _ = clean_and_categorize_part(pred, remove_dt)
    obj, triple_type = clean_and_categorize_part(obj, remove_dt)
    return subj, pred, obj, triple_type


class OAEI(CacheableEADataset):
    """The  OAEI (Ontology Alignment Evaluation Initiative) Knowledge Graph Track tasks contain graphs created from fandom wikis.

    Five integration tasks are available:
        - starwars-swg
        - starwars-swtor
        - marvelcinematicuniverse-marvel
        - memoryalpha-memorybeta
        - memoryalpha-stexpanded

    More information can be found at the `website <http://oaei.ontologymatching.org/2019/knowledgegraph/index.html>`_.
    """

    _TASK_URLS: Dict[OAEI_TASK_NAME, URL_SHA512_HASH] = {
        "starwars-swg": URL_SHA512_HASH(
            "https://cloud.scadsai.uni-leipzig.de/index.php/s/WzRxffgoq2qqn7q/download/starwars-swg.tar.gz",
            "ac6c1ca1b374e24fbeaf368a4ecf5254c54be38d0cbc1880815310ffa534c2266fa0da840014facf5fff430f777ef342a2d432fe00cf6fc14a0a12419d6966be",
        ),
        "starwars-swtor": URL_SHA512_HASH(
            "https://cloud.scadsai.uni-leipzig.de/index.php/s/CcG6tANX4JqacNY/download/starwars-swtor.tar.gz",
            "6b7913c5613bbc8c2aa849936d8b86c4613b73cebd06f8ee1c5e26f3605f95e61115a8a4d2d1399dbeb047579440899e9ecfce31f8c95c6013260dd3785cac80",
        ),
        "marvelcinematicuniverse-marvel": URL_SHA512_HASH(
            "https://cloud.scadsai.uni-leipzig.de/index.php/s/CH3NNin95BGjgb3/download/marvelcinematicuniverse-marvel.tar.gz",
            "d67936030c5ada48e6415774210527af51bf964828edd3f18beed51146ea9f70425cb760d073803189c4079dc68a4e39168fb1105c42fa6e82b790eb01e4a207",
        ),
        "memoryalpha-memorybeta": URL_SHA512_HASH(
            "https://cloud.scadsai.uni-leipzig.de/index.php/s/d6TYKFfDKxZEpiy/download/memoryalpha-memorybeta.tar.gz",
            "afb0a06225afcc2ae6edf5dcd513e48d4159e3d9bf94fc3cda6495a8c98a26fbc369413c587daec09e6ff7d045bf061a16ab8b65af4e22d0c91113390dc5d8d3",
        ),
        "memoryalpha-stexpanded": URL_SHA512_HASH(
            "https://cloud.scadsai.uni-leipzig.de/index.php/s/wSaRamAT5R5ieFZ/download/memoryalpha-stexpanded.tar.gz",
            "3d6f93c46425b4cb06f5a532601af9ee0a9fcdc3928709fa0ac66617099a27e26c244c024ae097274173cf2c910b045ea412fb225f51323eb7cdad86ad2461cc",
        ),
    }

    # avoid triggering dask compute
    _precalc_ds_statistics = {
        "starwars-swg": dict(
            rel_triples_left=6675247,
            rel_triples_right=178085,
            attr_triples_left=1570786,
            attr_triples_right=76269,
            ent_links=1096,
            folds=None,
            property_links=20,
            class_links=5,
        ),
        "starwars-swtor": dict(
            rel_triples_left=6675247,
            rel_triples_right=105543,
            attr_triples_left=1570786,
            attr_triples_right=40605,
            ent_links=1358,
            folds=None,
            property_links=56,
            class_links=15,
        ),
        "marvelcinematicuniverse-marvel": dict(
            rel_triples_left=1094598,
            rel_triples_right=5152898,
            attr_triples_left=130517,
            attr_triples_right=1580468,
            ent_links=1654,
            folds=None,
            property_links=11,
            class_links=2,
        ),
        "memoryalpha-memorybeta": dict(
            rel_triples_left=2096198,
            rel_triples_right=2048728,
            attr_triples_left=430730,
            attr_triples_right=494181,
            ent_links=9296,
            folds=None,
            property_links=55,
            class_links=14,
        ),
        "memoryalpha-stexpanded": dict(
            rel_triples_left=2096198,
            rel_triples_right=412179,
            attr_triples_left=430730,
            attr_triples_right=155207,
            ent_links=1725,
            folds=None,
            property_links=41,
            class_links=13,
        ),
    }

    def __init__(
        self,
        task: OAEI_TASK_NAME = "starwars-swg",
        backend: BACKEND_LITERAL = "dask",
        npartitions: int = 1,
        use_cache: bool = True,
        cache_path: Optional[pathlib.Path] = None,
    ):
        """Initialize a OAEI Knowledge Graph Track task.

        :param task: Name of the task. Has to be one of {starwars-swg,starwars-swtor,marvelcinematicuniverse-marvel,memoryalpha-memorybeta, memoryalpha-stexpanded}
        :param backend: Whether to use "pandas" or "dask"
        :param npartitions: how many partitions to use for each frame, when using dask
        :param use_cache: whether to use cache or not
        :param cache_path: Path where cache will be stored/loaded
        :raises ValueError: if unknown task value is provided
        """
        if task not in typing.get_args(OAEI_TASK_NAME):
            raise ValueError(f"Task has to be one of {OAEI_TASK_NAME}, but got{task}")
        self.task = task

        left_name, right_name = task.split("-")
        inner_cache_path = f"{left_name}_{right_name}"
        actual_cache_path = self.create_cache_path(
            OAEI_MODULE, inner_cache_path, cache_path
        )

        super().__init__(
            use_cache=use_cache,
            cache_path=actual_cache_path,
            dataset_names=(left_name, right_name),
            backend=backend,
            npartitions=npartitions,
        )

    def initial_read(self, backend: BACKEND_LITERAL):
        archive_url, sha512hash = OAEI._TASK_URLS[self.task]

        # ensure archive file is present
        archive_path = str(
            OAEI_MODULE.ensure(
                url=archive_url,
                download_kwargs=dict(hexdigests=dict(sha512=sha512hash)),
            )
        )
        entity_mapping_df, property_mapping_df, class_mapping_df = self._mapping_dfs(
            archive_path
        )
        left_attr, left_rel = self._read_triples(archive_path, left=True)
        right_attr, right_rel = self._read_triples(archive_path, left=False)

        # need to set before initialization
        # because of possible transforming
        self.property_links = property_mapping_df
        self.class_links = class_mapping_df
        return dict(
            rel_triples_left=left_rel,
            rel_triples_right=right_rel,
            attr_triples_left=left_attr,
            attr_triples_right=right_attr,
            ent_links=entity_mapping_df,
        )

    @property
    def _canonical_name(self) -> str:
        return f"{self.__class__.__name__}_{self.task}"

    @property
    def _param_repr(self) -> str:
        return f"task={self.task}, "

    def _additional_backend_handling(self, backend: BACKEND_LITERAL):
        if backend == "pandas":
            self._backend = "pandas"
            if isinstance(self.property_links, pd.DataFrame):
                return
            else:
                self.property_links = self.property_links.compute()
                self.class_links = self.class_links.compute()
        elif backend == "dask":
            self._backend = "dask"
            if isinstance(self.property_links, dd.DataFrame):
                if self.property_links.npartitions != self.npartitions:
                    self.property_links = self.property_links.repartition(
                        npartitions=self.npartitions
                    )
                    self.class_links = self.class_links.repartition(
                        npartitions=self.npartitions
                    )
            else:
                self.property_links = dd.from_pandas(
                    self.property_links, npartitions=self.npartitions
                )
                self.class_links = dd.from_pandas(
                    self.class_links, npartitions=self.npartitions
                )

    @property
    def _statistics(self) -> str:
        this_ds_stats = self.__class__._precalc_ds_statistics[self.task]
        stat_str = ""
        for key, value in this_ds_stats.items():
            stat_str += f"{key}={value}, "
        return stat_str[:-2]  # remove last comma and space

    def _pair_dfs(self, mapping_df: dd.DataFrame) -> dd.DataFrame:
        left_side_uri = (
            "http://knowledgeweb.semanticweb.org/heterogeneity/alignmententity1"
        )
        side_predicate = mapping_df[LABEL_RELATION] == left_side_uri
        map_left = mapping_df[side_predicate]
        map_right = mapping_df[~side_predicate]
        joined_left = LABEL_TAIL + EA_SIDE_LEFT
        joined_right = LABEL_TAIL + EA_SIDE_RIGHT
        return (
            map_left.set_index(LABEL_HEAD)
            .join(
                map_right.set_index(LABEL_HEAD),
                lsuffix=EA_SIDE_LEFT,
                rsuffix=EA_SIDE_RIGHT,
            )[[joined_left, joined_right]]
            .rename(columns={joined_left: EA_SIDE_LEFT, joined_right: EA_SIDE_RIGHT})
        )

    def _mapping_dfs(
        self, archive_path: str
    ) -> Tuple[dd.DataFrame, dd.DataFrame, dd.DataFrame]:
        LINE_TYPE_COL = "line_type"
        schema_df = pd.DataFrame(
            {
                LABEL_HEAD: pd.Series([], dtype="str"),
                LABEL_RELATION: pd.Series([], dtype="str"),
                LABEL_TAIL: pd.Series([], dtype="str"),
                LINE_TYPE_COL: pd.Series(
                    [],
                    dtype=reflinetype,
                ),
            }
        )
        df = (
            read_dask_bag_from_archive_text(
                archive_path,
                inner_path=f"{self.task}/ref-{self.task}.nt",
                protocol="tar",
            )
            .map(parse_ref_line)
            .filter(lambda x: x[3] != "dismiss")
            .to_dataframe(schema_df)
        )

        entity_mapping_df = df[df[LINE_TYPE_COL] == "entity"]
        property_mapping_df = df[df[LINE_TYPE_COL] == "property"]
        class_mapping_df = df[df[LINE_TYPE_COL] == "class"]
        return (
            self._pair_dfs(entity_mapping_df),
            self._pair_dfs(property_mapping_df),
            self._pair_dfs(class_mapping_df),
        )

    def _read_triples(
        self, archive_path: str, left: bool
    ) -> Tuple[dd.DataFrame, dd.DataFrame]:
        task_index = 0 if left else 1
        task_side = self.task.split("-")[task_index]
        REL_ATTR_COL = "rel_attr_type"
        schema_df = pd.DataFrame(
            {
                LABEL_HEAD: pd.Series([], dtype="str"),
                LABEL_RELATION: pd.Series([], dtype="str"),
                LABEL_TAIL: pd.Series([], dtype="str"),
                REL_ATTR_COL: pd.Series(
                    [],
                    dtype=relattrlinetype,
                ),
            }
        )
        df = (
            read_dask_bag_from_archive_text(
                archive_path, inner_path=f"{self.task}/{task_side}.nt", protocol="tar"
            )
            .map(fault_tolerant_parse_nt)
            .to_dataframe(schema_df)
        )
        return (
            df[df[REL_ATTR_COL] == "attr"][COLUMNS],
            df[df[REL_ATTR_COL] == "rel"][COLUMNS],
        )



from typing import TYPE_CHECKING, Any, Mapping, Optional

if TYPE_CHECKING:
    import dask.bag as db
    import dask.dataframe as dd


def read_dask_bag_from_archive_text(
    path: str,
    inner_path: str,
    protocol: str,
    read_text_kwargs: Optional[Mapping[str, Any]] = None,
) -> "db.Bag":
    import dask.bag as db

    actual_path = f"{protocol}://{inner_path}::{path}"
    text_kwargs = read_text_kwargs or {}
    return db.read_text(actual_path, **text_kwargs)


def read_dask_df_archive_csv(
    path: str,
    inner_path: str,
    protocol: str,
    **read_csv_kwargs,
) -> "dd.DataFrame":
    import dask.dataframe as dd

    actual_path = f"{protocol}://{inner_path}::{path}"
    return dd.read_csv(actual_path, **read_csv_kwargs)



import logging
from typing import Type

logger = logging.getLogger(__name__)


def fix_dataclass_init_docs(cls: Type) -> Type:
    """Fix the ``__init__`` documentation for a :class:`dataclasses.dataclass`.
    :param cls: The class whose docstring needs fixing
    :returns: The class that was passed so this function can be used as a decorator

    .. seealso:: https://github.com/agronholm/sphinx-autodoc-typehints/issues/123
    """
    cls.__init__.__qualname__ = f"{cls.__name__}.__init__"
    return cls



import pathlib
from typing import Literal, Optional, Tuple

from moviegraphbenchmark import load_data

from .base import (
    BACKEND_LITERAL,
    BASE_DATASET_MODULE,
    CacheableEADataset,
    TrainTestValSplit,
)

MOVIEGRAPH_MODULE = BASE_DATASET_MODULE.module("moviegraphbenchmark")

# graph pairs
GraphPair = Literal["imdb-tmdb", "imdb-tvdb", "tmdb-tvdb"]
IMDB_TMDB: GraphPair = "imdb-tmdb"
IMDB_TVDB: GraphPair = "imdb-tvdb"
TMDB_TVDB: GraphPair = "tmdb-tvdb"

GRAPH_PAIRS: Tuple[GraphPair, ...] = (IMDB_TMDB, IMDB_TVDB, TMDB_TVDB)


class MovieGraphBenchmark(CacheableEADataset):
    """Class containing the movie graph benchmark.
    Published in `Obraczka, D. et. al. (2021) Embedding-Assisted Entity Resolution for Knowledge Graphs <http://ceur-ws.org/Vol-2873/paper8.pdf>`_,
    *Proceedings of the 2nd International Workshop on Knowledge Graph Construction co-located with 18th Extended Semantic Web Conference*"""

    def __init__(
        self,
        graph_pair: GraphPair = "imdb-tmdb",
        backend: BACKEND_LITERAL = "pandas",
        npartitions: int = 1,
        use_cache: bool = True,
        cache_path: Optional[pathlib.Path] = None,
    ):
        """Initialize a MovieGraphBenchmark dataset.

        :param graph_pair: which graph pair to use of "imdb-tdmb","imdb-tvdb" or "tmdb-tvdb"
        :param backend: Whether to use "pandas" or "dask"
        :param npartitions: how many partitions to use for each frame, when using dask
        :param use_cache: whether to use cache or not
        :param cache_path: Path where cache will be stored/loaded
        :raises ValueError: if unknown graph pair
        """
        # Input validation.
        if graph_pair not in GRAPH_PAIRS:
            raise ValueError(f"Invalid graph pair: Allowed are: {GRAPH_PAIRS}")

        self.graph_pair = graph_pair

        actual_cache_path = self.create_cache_path(
            MOVIEGRAPH_MODULE, graph_pair, cache_path
        )
        left_name, right_name = self.graph_pair.split("-")
        super().__init__(
            cache_path=actual_cache_path,
            use_cache=use_cache,
            backend=backend,
            npartitions=npartitions,
            dataset_names=(left_name, right_name),
        )

    def initial_read(self, backend: BACKEND_LITERAL):
        ds = load_data(pair=self.graph_pair, data_path=str(MOVIEGRAPH_MODULE.base))
        folds = [
            TrainTestValSplit(
                train=fold.train_links, test=fold.test_links, val=fold.valid_links
            )
            for fold in ds.folds
        ]
        return dict(
            rel_triples_left=ds.rel_triples_1,
            rel_triples_right=ds.rel_triples_2,
            attr_triples_left=ds.attr_triples_1,
            attr_triples_right=ds.attr_triples_2,
            ent_links=ds.ent_links,
            folds=folds,
        )

    @property
    def _canonical_name(self) -> str:
        return f"{self.__class__.__name__}_{self.graph_pair}"

    @property
    def _param_repr(self) -> str:
        return f"graph_pair={self.graph_pair}, "



from typing import Literal, Tuple

# borrowed from pykeen.typing
Target = Literal["head", "relation", "tail"]
LABEL_HEAD: Target = "head"
LABEL_RELATION: Target = "relation"
LABEL_TAIL: Target = "tail"
EASide = Literal["left", "right"]
EA_SIDE_LEFT: EASide = "left"
EA_SIDE_RIGHT: EASide = "right"
EA_SIDES: Tuple[EASide, EASide] = (EA_SIDE_LEFT, EA_SIDE_RIGHT)
COLUMNS = [LABEL_HEAD, LABEL_RELATION, LABEL_TAIL]
BACKEND_LITERAL = Literal["pandas", "dask"]



from dataclasses import dataclass
from typing import Dict, Iterable, Mapping, Optional, Sequence, Tuple

import numpy as np
import pandas as pd

from .base import EADataset, TrainTestValSplit
from .utils import fix_dataclass_init_docs


def enhance_mapping(
    labels: Iterable, mapping: Mapping[str, int] = None
) -> Dict[str, int]:
    """Map labels with given mapping and enhance mapping if unseen labels are encountered.

    :param labels: Labels to map
    :param mapping: Know mappings
    :return: Enhanced mapping
    """
    mapping = {} if mapping is None else mapping
    new_id = len(mapping)
    enhanced_mapping = {**mapping}
    for label in labels:
        label = str(label)
        if label not in mapping:
            enhanced_mapping[label] = new_id
            new_id += 1
    return enhanced_mapping


def perform_map(
    triples: np.ndarray,
    head_map: Mapping[str, int],
    rel_map: Mapping[str, int],
    tail_map: Mapping[str, int],
) -> np.ndarray:
    """Map str triples to int ids via dictionaries.

    :param triples: string triples
    :param head_map: mapping for head column
    :param rel_map: mapping for rel column
    :param tail_map: mapping for tail column
    :return: integer id mapped triples
    """
    head_getter = np.vectorize(head_map.get)
    rel_getter = np.vectorize(rel_map.get)
    tail_getter = np.vectorize(tail_map.get)
    # have to use triples with start:end instead of heads/rels/tails variable
    # because this way we get array of shape (n,1) instead of (n,)
    head_column = head_getter(triples[:, 0:1])
    rel_column = rel_getter(triples[:, 1:2])
    tail_column = tail_getter(triples[:, 2:3])
    return np.concatenate([head_column, rel_column, tail_column], axis=1)


def id_map_rel_triples(
    df: pd.DataFrame,
    entity_mapping: Dict[str, int] = None,
    rel_mapping: Dict[str, int] = None,
) -> Tuple[np.ndarray, Dict[str, int], Dict[str, int]]:
    """Map entity and relation labels to ids and create numpy array.

    :param df: labeled triples
    :param entity_mapping: already mapped entities
    :param rel_mapping: already mapped relations
    :return: id-based numpy array triples, (updated) entity label to id mapping, (updated) relation label to id mapping
    """
    triples = df.astype(str).values
    heads, rels, tails = triples[:, 0], triples[:, 1], triples[:, 2]
    # sorting  ensures consistent results
    entity_labels = sorted(set(heads).union(tails))
    relation_labels = sorted(set(rels))
    entity_mapping = enhance_mapping(entity_labels, entity_mapping)
    rel_mapping = enhance_mapping(relation_labels, rel_mapping)
    return (
        perform_map(triples, entity_mapping, rel_mapping, entity_mapping),
        entity_mapping,
        rel_mapping,
    )


def _id_map_attr_triples(
    df: pd.DataFrame,
    entity_mapping: Dict[str, int],
    attr_rel_mapping: Dict[str, int] = None,
    attr_mapping: Dict[str, int] = None,
) -> Tuple[np.ndarray, Dict[str, int], Dict[str, int], Dict[str, int]]:
    """Map entity, relation labels and attributes to ids and create numpy array.

    :param df: labeled triples
    :param entity_mapping: already mapped entities
    :param attr_rel_mapping: already mapped attribute relations
    :param attr_mapping: already mapped attributes
    :return: id-based numpy array triples, (updated) entity label to id mapping, (updated) relation label to id mapping, (updated) attribute to id mapping
    """
    triples = df.astype(str).values
    heads, rels, tails = triples[:, 0], triples[:, 1], triples[:, 2]
    # sorting  ensures consistent results
    entity_labels = sorted(set(heads))
    relation_labels = sorted(set(rels))
    attributes = sorted(set(tails))
    entity_mapping = enhance_mapping(entity_labels, entity_mapping)
    rel_mapping = enhance_mapping(relation_labels, attr_rel_mapping)
    attr_mapping = enhance_mapping(attributes, attr_mapping)
    return (
        perform_map(triples, entity_mapping, rel_mapping, attr_mapping),
        entity_mapping,
        rel_mapping,
        attr_mapping,
    )


def _map_links(links: pd.DataFrame, entity_mapping: Dict[str, int]) -> np.ndarray:
    """Map links via given mapping.

    :param links: entity links
    :param entity_mapping: label to id mapping
    :return: numpy array with ids
    """
    tuples = links.values
    entity_getter = np.vectorize(entity_mapping.get)
    return np.concatenate(
        [entity_getter(tuples[:, 0:1]), entity_getter(tuples[:, 1:2])], axis=1
    )


@fix_dataclass_init_docs
@dataclass
class IdMappedTrainTestValSplit:
    """Dataclass holding split of gold standard entity links."""

    #: entity links for training
    train: np.ndarray
    #: entity links for testing
    test: np.ndarray
    #: entity links for validation
    val: np.ndarray


@fix_dataclass_init_docs
@dataclass
class IdMappedEADataset:
    """Dataclass holding information of the alignment class with mapping of string to numerical id."""

    #: relation triples of left knowledge graph
    rel_triples_left: np.ndarray
    #: relation triples of right knowledge graph
    rel_triples_right: np.ndarray
    #: attribute triples of left knowledge graph
    attr_triples_left: np.ndarray
    #: attribute triples of right knowledge graph
    attr_triples_right: np.ndarray
    #: gold standard entity links of alignment
    ent_links: np.ndarray
    #: label to id mapping for all entities
    entity_mapping: Dict[str, int]
    #: label to id mapping for all relations
    rel_mapping: Dict[str, int]
    #: label to id mapping for all attribute relations
    attr_rel_mapping: Dict[str, int]
    #: attribute to id mapping for all attributes
    attr_mapping: Dict[str, int]
    #: optional pre-split folds of the gold standard
    folds: Optional[Sequence[IdMappedTrainTestValSplit]] = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(rel_triples_left={len(self.rel_triples_left)}, rel_triples_right={len(self.rel_triples_right)}, attr_triples_left={len(self.attr_triples_left)}, attr_triples_right={len(self.attr_triples_right)}, ent_links={len(self.ent_links)}, entity_mapping={len(self.entity_mapping)}, rel_mapping={len(self.rel_mapping)}, attr_rel_mapping={len(self.attr_rel_mapping)}, attr_mapping={len(self.attr_mapping)}, folds={len(self.folds) if self.folds else None})"

    @classmethod
    def from_frames(
        cls,
        rel_triples_left: pd.DataFrame,
        rel_triples_right: pd.DataFrame,
        attr_triples_left: pd.DataFrame,
        attr_triples_right: pd.DataFrame,
        ent_links: pd.DataFrame,
        folds: Optional[Sequence[TrainTestValSplit]],
    ) -> "IdMappedEADataset":
        rel_triples_left, entity_mapping, rel_mapping = id_map_rel_triples(
            rel_triples_left
        )
        rel_triples_right, entity_mapping, rel_mapping = id_map_rel_triples(
            rel_triples_right,
            entity_mapping=entity_mapping,
            rel_mapping=rel_mapping,
        )
        (
            attr_triples_left,
            entity_mapping,
            attr_rel_mapping,
            attr_mapping,
        ) = _id_map_attr_triples(attr_triples_left, entity_mapping=entity_mapping)
        (
            attr_triples_right,
            entity_mapping,
            attr_rel_mapping,
            attr_mapping,
        ) = _id_map_attr_triples(
            attr_triples_right,
            entity_mapping=entity_mapping,
            attr_rel_mapping=attr_rel_mapping,
            attr_mapping=attr_mapping,
        )

        ent_links = _map_links(ent_links, entity_mapping)
        new_folds = None
        if folds:
            new_folds = []
            for fold in folds:
                train = _map_links(fold.train, entity_mapping)
                test = _map_links(fold.test, entity_mapping)
                val = _map_links(fold.val, entity_mapping)
                new_folds.append(
                    IdMappedTrainTestValSplit(train=train, test=test, val=val)
                )
        return cls(
            rel_triples_left=rel_triples_left,
            rel_triples_right=rel_triples_right,
            attr_triples_left=attr_triples_left,
            attr_triples_right=attr_triples_right,
            ent_links=ent_links,
            entity_mapping=entity_mapping,
            rel_mapping=rel_mapping,
            attr_rel_mapping=attr_rel_mapping,
            attr_mapping=attr_mapping,
            folds=new_folds,
        )

    @classmethod
    def from_ea_dataset(cls, dataset: EADataset) -> "IdMappedEADataset":
        return IdMappedEADataset.from_frames(
            rel_triples_left=dataset.rel_triples_left,
            rel_triples_right=dataset.rel_triples_right,
            attr_triples_left=dataset.attr_triples_left,
            attr_triples_right=dataset.attr_triples_right,
            ent_links=dataset.ent_links,
            folds=dataset.folds,
        )



import logging
from importlib.metadata import version  # pragma: no cover

from .base import EADataset
from .id_mapped import IdMappedEADataset
from .moviegraph_benchmark_loader import MovieGraphBenchmark
from .oaei_loader import OAEI
from .open_ea_loader import OpenEA

__all__ = [
    "OpenEA",
    "MovieGraphBenchmark",
    "OAEI",
    "IdMappedEADataset",
    "EADataset",
]
__version__ = version(__package__)
logging.getLogger(__name__).setLevel(logging.INFO)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sylloge']

package_data = \
{'': ['*']}

install_requires = \
['dask>=2022.01.0',
 'moviegraphbenchmark>=1.0.1,<2.0.0',
 'pandas>=1.0',
 'pyarrow',
 'pystow>=0.4.6,<0.5.0',
 'python-slugify>=7.0.0']

extras_require = \
{'docs': ['Sphinx>=5.0.0,<6.0.0',
          'insegel>=1.3.1,<2.0.0',
          'sphinx-automodapi>=0.14.1,<0.15.0',
          'sphinx-autodoc-typehints>=1.19.2,<2.0.0']}

setup_kwargs = {
    'name': 'sylloge',
    'version': '0.2.0',
    'description': 'Small library to simplify collecting and loading of entity alignment benchmark datasets',
    'long_description': '<p align="center">\n<img src="https://github.com/dobraczka/sylloge/raw/main/docs/logo.png" alt="sylloge logo", width=200/>\n</p>\n\n<h2 align="center">sylloge</h2>\n\n<p align="center">\n<a href="https://github.com/dobraczka/sylloge/actions/workflows/main.yml"><img alt="Actions Status" src="https://github.com/dobraczka/sylloge/actions/workflows/main.yml/badge.svg?branch=main"></a>\n<a href=\'https://sylloge.readthedocs.io/en/latest/?badge=latest\'><img src=\'https://readthedocs.org/projects/sylloge/badge/?version=latest\' alt=\'Documentation Status\' /></a>\n<a href="https://pypi.org/project/sylloge"/><img alt="Stable python versions" src="https://img.shields.io/pypi/pyversions/sylloge"></a>\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n</p>\n\nThis simple library aims to collect entity-alignment benchmark datasets and make them easily available.\n\nUsage\n=====\nLoad benchmark datasets:\n```\n>>> from sylloge import OpenEA\n>>> ds = OpenEA()\n>>> ds\nOpenEA(backend=pandas, graph_pair=D_W, size=15K, version=V1, rel_triples_left=38265, rel_triples_right=42746, attr_triples_left=52134, attr_triples_right=138246, ent_links=15000, folds=5)\n>>> ds.rel_triples_right.head()\n                                       head                             relation                                    tail\n0   http://www.wikidata.org/entity/Q6176218   http://www.wikidata.org/entity/P27     http://www.wikidata.org/entity/Q145\n1   http://www.wikidata.org/entity/Q212675  http://www.wikidata.org/entity/P161  http://www.wikidata.org/entity/Q446064\n2   http://www.wikidata.org/entity/Q13512243  http://www.wikidata.org/entity/P840      http://www.wikidata.org/entity/Q84\n3   http://www.wikidata.org/entity/Q2268591   http://www.wikidata.org/entity/P31   http://www.wikidata.org/entity/Q11424\n4   http://www.wikidata.org/entity/Q11300470  http://www.wikidata.org/entity/P178  http://www.wikidata.org/entity/Q170420\n>>> ds.attr_triples_left.head()\n                                  head                                          relation                                               tail\n0  http://dbpedia.org/resource/E534644                http://dbpedia.org/ontology/imdbId                                            0044475\n1  http://dbpedia.org/resource/E340590               http://dbpedia.org/ontology/runtime  6480.0^^<http://www.w3.org/2001/XMLSchema#double>\n2  http://dbpedia.org/resource/E840454  http://dbpedia.org/ontology/activeYearsStartYear     1948^^<http://www.w3.org/2001/XMLSchema#gYear>\n3  http://dbpedia.org/resource/E971710       http://purl.org/dc/elements/1.1/description                          English singer-songwriter\n4  http://dbpedia.org/resource/E022831       http://dbpedia.org/ontology/militaryCommand                     Commandant of the Marine Corps\n>>> ds.ent_links.head()\n                                  left                                    right\n0  http://dbpedia.org/resource/E123186    http://www.wikidata.org/entity/Q21197\n1  http://dbpedia.org/resource/E228902  http://www.wikidata.org/entity/Q5909974\n2  http://dbpedia.org/resource/E718575   http://www.wikidata.org/entity/Q707008\n3  http://dbpedia.org/resource/E469216  http://www.wikidata.org/entity/Q1471945\n4  http://dbpedia.org/resource/E649433  http://www.wikidata.org/entity/Q1198381\n```\n\nYou can get a canonical name for a dataset instance to use e.g. to create folders to store experiment results:\n\n```\n   >>> ds.canonical_name\n   \'openea_d_w_15k_v1\'\n```\n\nCreate id-mapped dataset for embedding-based methods:\n\n```\n>>> from sylloge import IdMappedEADataset\n>>> id_mapped_ds = IdMappedEADataset.from_ea_dataset(ds)\n>>> id_mapped_ds\nIdMappedEADataset(rel_triples_left=38265, rel_triples_right=42746, attr_triples_left=52134, attr_triples_right=138246, ent_links=15000, entity_mapping=30000, rel_mapping=417, attr_rel_mapping=990, attr_mapping=138836, folds=5)\n>>> id_mapped_ds.rel_triples_right\n[[26048   330 16880]\n [19094   293 23348]\n [16554   407 29192]\n ...\n [16480   330 15109]\n [18465   254 19956]\n [26040   290 28560]]\n```\n\nYou can use [dask](https://www.dask.org/) as backend for larger datasets:\n```\n>>> ds = OpenEA(backend="dask")\n>>> ds\nOpenEA(backend=dask, graph_pair=D_W, size=15K, version=V1, rel_triples_left=38265, rel_triples_right=42746, attr_triples_left=52134, attr_triples_right=138246, ent_links=15000, folds=5)\n```\nWhich replaces pandas DataFrames with dask DataFrames.\n\nDatasets can be written/read as parquet via `to_parquet` or `read_parquet`.\nAfter the initial read datasets are cached using this format. The `cache_path` can be explicitly set and caching behaviour can be disable via `use_cache=False`, when initalizing a dataset.\n\nSome datasets come with pre-determined splits:\n\n```bash\ntree ~/.data/sylloge/open_ea/cached/D_W_15K_V1 \n├── attr_triples_left_parquet\n├── attr_triples_right_parquet\n├── dataset_names.txt\n├── ent_links_parquet\n├── folds\n│\xa0\xa0 ├── 1\n│\xa0\xa0 │\xa0\xa0 ├── test_parquet\n│\xa0\xa0 │\xa0\xa0 ├── train_parquet\n│\xa0\xa0 │\xa0\xa0 └── val_parquet\n│\xa0\xa0 ├── 2\n│\xa0\xa0 │\xa0\xa0 ├── test_parquet\n│\xa0\xa0 │\xa0\xa0 ├── train_parquet\n│\xa0\xa0 │\xa0\xa0 └── val_parquet\n│\xa0\xa0 ├── 3\n│\xa0\xa0 │\xa0\xa0 ├── test_parquet\n│\xa0\xa0 │\xa0\xa0 ├── train_parquet\n│\xa0\xa0 │\xa0\xa0 └── val_parquet\n│\xa0\xa0 ├── 4\n│\xa0\xa0 │\xa0\xa0 ├── test_parquet\n│\xa0\xa0 │\xa0\xa0 ├── train_parquet\n│\xa0\xa0 │\xa0\xa0 └── val_parquet\n│\xa0\xa0 └── 5\n│\xa0\xa0     ├── test_parquet\n│\xa0\xa0     ├── train_parquet\n│\xa0\xa0     └── val_parquet\n├── rel_triples_left_parquet\n└── rel_triples_right_parquet\n```\nsome don\'t:\n```bash\ntree ~/.data/sylloge/oaei/cached/starwars_swg\n├── attr_triples_left_parquet\n│\xa0\xa0 └── part.0.parquet\n├── attr_triples_right_parquet\n│\xa0\xa0 └── part.0.parquet\n├── dataset_names.txt\n├── ent_links_parquet\n│\xa0\xa0 └── part.0.parquet\n├── rel_triples_left_parquet\n│\xa0\xa0 └── part.0.parquet\n└── rel_triples_right_parquet\n    └── part.0.parquet\n```\n\n\nInstallation\n============\n```bash\npip install sylloge \n```\n\nDatasets\n========\n| Dataset family name | Year | # of Datasets | Sources | References |\n|:--------------------|:----:|:-------------:|:-------:|:----------|\n| [OpenEA](https://sylloge.readthedocs.io/en/latest/source/datasets.html#sylloge.OpenEA) | 2020 | 16 | DBpedia, Yago, Wikidata |  [Paper](http://www.vldb.org/pvldb/vol13/p2326-sun.pdf), [Repo](https://github.com/nju-websoft/OpenEA#dataset-overview) |\n| [MovieGraphBenchmark](https://sylloge.readthedocs.io/en/latest/source/datasets.html#sylloge.MovieGraphBenchmark) | 2022 | 3 | IMDB, TMDB, TheTVDB | [Paper](http://ceur-ws.org/Vol-2873/paper8.pdf), [Repo](https://github.com/ScaDS/MovieGraphBenchmark) |\n| [OAEI](https://sylloge.readthedocs.io/en/latest/source/datasets.html#sylloge.OAEI) | 2022 | 5 | Fandom wikis | [Paper](https://ceur-ws.org/Vol-3324/oaei22_paper0.pdf), [Website](http://oaei.ontologymatching.org/2022/knowledgegraph/index.html) |\n',
    'author': 'Daniel Obraczka',
    'author_email': 'obraczka@informatik.uni-leipzig.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/dobraczka/sylloge',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# largely adapted from pykeen.datasets.ea.openea
import pathlib
from typing import Dict, Literal, Optional, Tuple

from .base import BACKEND_LITERAL, BASE_DATASET_MODULE, ZipEADatasetWithPreSplitFolds

OPEN_EA_MODULE = BASE_DATASET_MODULE.module("open_ea")

# graph pairs
GraphPair = Literal["D_W", "D_Y", "EN_DE", "EN_FR"]
D_W: GraphPair = "D_W"
D_Y: GraphPair = "D_Y"
EN_DE: GraphPair = "EN_DE"
EN_FR: GraphPair = "EN_FR"
GRAPH_PAIRS: Tuple[GraphPair, ...] = (D_W, D_Y, EN_DE, EN_FR)

# graph sizes
GraphSize = Literal["15K", "100K"]
SIZE_15K: GraphSize = "15K"
SIZE_100K: GraphSize = "100K"
GRAPH_SIZES = (SIZE_15K, SIZE_100K)

# graph versions
GraphVersion = Literal["V1", "V2"]
V1: GraphVersion = "V1"
V2: GraphVersion = "V2"
GRAPH_VERSIONS = (V1, V2)


class OpenEA(ZipEADatasetWithPreSplitFolds):
    """Class containing the OpenEA dataset family.
    Published in `Sun, Z. et. al. (2020) A Benchmarking Study of Embedding-based Entity Alignment for Knowledge Graphs <http://www.vldb.org/pvldb/vol13/p2326-sun.pdf>`_,
    *Proceedings of the VLDB Endowment*"""

    #: The link to the zip file
    _FIGSHARE_LINK: str = "https://figshare.com/ndownloader/files/34234391"

    #: The hex digest for the zip file
    _SHA512: str = (
        "c1589f185f86e05c497de147b4d6c243c66775cb4b50c6b41ecc71b36cfafb4c"
        "9f86fbee94e1e78a7ee056dd69df1ce3fc210ae07dc64955ad2bfda7450545ef"
    )

    _GRAPH_PAIR_TO_DS_NAMES: Dict[GraphPair, Tuple[str, str]] = {
        "D_W": ("DBpedia", "Wikidata"),
        "D_Y": ("DBpedia", "YAGO"),
        "EN_DE": ("DBpedia_EN", "DBpedia_DE"),
        "EN_FR": ("DBpedia_EN", "DBpedia_FR"),
    }

    def __init__(
        self,
        graph_pair: GraphPair = "D_W",
        size: GraphSize = "15K",
        version: GraphVersion = "V1",
        backend: BACKEND_LITERAL = "pandas",
        npartitions: int = 1,
        use_cache: bool = True,
        cache_path: Optional[pathlib.Path] = None,
    ):
        """Initializes an OpenEA dataset.

        :param graph_pair: which pair to use of "D_W", "D_Y", "EN_DE" or "EN_FR"
        :param size: what size ("15K" or "100K")
        :param version: which version to use ("V1" or "V2")
        :param backend: Whether to use "pandas" or "dask"
        :param npartitions: how many partitions to use for each frame, when using dask
        :param use_cache: whether to use cache or not
        :param cache_path: Path where cache will be stored/loaded
        :raises ValueError: if unknown graph_pair,size or version values are provided
        """
        # Input validation.
        if graph_pair not in GRAPH_PAIRS:
            raise ValueError(f"Invalid graph pair: Allowed are: {GRAPH_PAIRS}")
        if size not in GRAPH_SIZES:
            raise ValueError(f"size must be one of {GRAPH_SIZES}")
        if version not in GRAPH_VERSIONS:
            raise ValueError(f"version must be one of {GRAPH_VERSIONS}")

        self.graph_pair = graph_pair
        self.size = size
        self.version = version

        # ensure zip file is present
        zip_path = OPEN_EA_MODULE.ensure(
            url=OpenEA._FIGSHARE_LINK,
            name="OpenEA_dataset_v2.0.zip",
            download_kwargs=dict(hexdigests=dict(sha512=OpenEA._SHA512)),
        )

        inner_cache_path = f"{graph_pair}_{size}_{version}"
        inner_path = pathlib.PurePosixPath("OpenEA_dataset_v2.0", inner_cache_path)
        actual_cache_path = self.create_cache_path(
            OPEN_EA_MODULE, inner_cache_path, cache_path
        )
        super().__init__(
            cache_path=actual_cache_path,
            use_cache=use_cache,
            zip_path=zip_path,
            inner_path=inner_path,
            backend=backend,
            npartitions=npartitions,
            dataset_names=OpenEA._GRAPH_PAIR_TO_DS_NAMES[graph_pair],
        )

    @property
    def _canonical_name(self) -> str:
        return f"{self.__class__.__name__}_{self.graph_pair}_{self.size}_{self.version}"

    @property
    def _param_repr(self) -> str:
        return (
            f"graph_pair={self.graph_pair}, size={self.size}, version={self.version}, "
        )
